import numpy as np
import pytest

import pipelime.sequences as pls


def _np_eq(x, y) -> bool:
    return np.array_equal(x, y, equal_nan=True)


class TestSample:
    def _np_sample(self):
        import pipelime.items as pli

        data = {k: pli.NpyNumpyItem(np.random.rand(3, 4)) for k in ("a", "b", "c")}
        return pls.Sample(data), data

    def _mixed_sample(self):
        import pipelime.items as pli

        data = {
            "c": pli.JsonMetadataItem({"foo": "bar"}),
            "d": pli.NpyNumpyItem(np.random.rand(3, 4)),
            "e": pli.TxtNumpyItem(np.random.rand(4)),
        }
        return pls.Sample(data), data

    def test_creation(self):
        sample, data = self._np_sample()

        assert len(sample) == len(data)
        assert all(k in sample for k in data)
        assert all(_np_eq(sample[k](), v()) for k, v in data.items())
        assert sample.to_dict() == {k: v() for k, v in data.items()}

    def test_to_schema(self):
        from pydantic import BaseConfig, Extra, create_model
        from typing import Optional

        import pipelime.items as pli

        class SampleConfig(BaseConfig):
            arbitrary_types_allowed = True
            extra = Extra.forbid

        sample, _ = self._mixed_sample()
        sample_schema = create_model(
            "SampleSchema",
            __config__=SampleConfig,
            **{k: (v, ...) for k, v in sample.to_schema().items()},
        )

        try:
            sample_schema(**sample)
        except Exception:
            assert False

        sample_schema = create_model(
            "SampleSchema",
            __config__=SampleConfig,
            c=(pli.MetadataItem, ...),
            d=(pli.NumpyItem, ...),
            e=(pli.Item, ...),
        )

        try:
            sample_schema(**sample)
        except Exception:
            assert False

        sample_schema = create_model(
            "SampleSchema",
            __config__=SampleConfig,
            c=(pli.MetadataItem, ...),
            d=(pli.NumpyItem, ...),
            e=(pli.ImageItem, ...),
        )

        with pytest.raises(Exception):
            sample_schema(**sample)

        class SampleConfig2(BaseConfig):
            arbitrary_types_allowed = True

        sample_schema = create_model(
            "SampleSchema",
            __config__=SampleConfig2,
            c=(pli.MetadataItem, ...),
            d=(pli.NumpyItem, ...),
            e_other=(Optional[pli.Item], None),
        )

        try:
            sample_schema(**sample)
        except Exception:
            assert False

    def test_shallow_copy(self):
        sample, _ = self._np_sample()
        other_sample = sample.shallow_copy()
        assert sample.keys() == other_sample.keys()
        assert all(v is other_sample[k] for k, v in sample.items())

    def test_deep_copy(self):
        sample, _ = self._np_sample()
        other_sample = sample.deep_copy()
        assert sample.keys() == other_sample.keys()
        assert all(v is not other_sample[k] for k, v in sample.items())

    def test_set_item(self):
        import pipelime.items as pli

        sample, data = self._np_sample()
        changed_key = next(iter(data))
        changed_item = pli.JsonMetadataItem({"a": 1})
        other_sample = sample.set_item(changed_key, changed_item)
        assert sample.keys() == other_sample.keys()
        assert all(
            (v is other_sample[k]) if k != changed_key else (v is not other_sample[k])
            for k, v in sample.items()
        )
        assert other_sample[changed_key] is changed_item

    def test_set_value_as(self):
        import pipelime.items as pli

        sample, data = self._mixed_sample()
        ref_key = next(k for k, v in data.items() if isinstance(v, pli.NumpyItem))
        target_key = "new_key"
        new_value = np.random.rand(4, 4)

        other_sample = sample.set_value_as(target_key, ref_key, new_value)

        assert all(k in other_sample for k in sample)
        assert all(k in sample for k in other_sample if k != target_key)
        assert all(
            (v is sample[k])
            if k != target_key
            else isinstance(v, sample[ref_key].__class__)
            for k, v in other_sample.items()
        )
        assert _np_eq(other_sample[target_key](), new_value)

    def test_set_value(self):
        import pipelime.items as pli

        sample, data = self._mixed_sample()
        target_key = next(k for k, v in data.items() if isinstance(v, pli.NumpyItem))
        new_value = np.random.rand(4, 4)

        other_sample = sample.set_value(target_key, new_value)

        assert sample.keys() == other_sample.keys()
        assert all(
            (v is sample[k])
            if k != target_key
            else isinstance(v, sample[target_key].__class__)
            for k, v in other_sample.items()
        )
        assert _np_eq(other_sample[target_key](), new_value)

    def test_deep_set(self):
        import pipelime.items as pli

        data = {
            "fir.st": pli.JsonMetadataItem(
                {"data": ["a", "b", {"c": 42}], "meta": 3.14}
            ),
            r"sec\ond": pli.JsonMetadataItem(["a", "b", {"c": 36}]),
        }
        sample = pls.Sample(data)

        other_sample = sample.deep_set(r"fir\.st.data[2].c", 17)
        assert sample["fir.st"]()["data"][2]["c"] == 42  # type: ignore
        assert other_sample["fir.st"]()["data"][2]["c"] == 17  # type: ignore

        other_sample = sample.deep_set(r"sec\ond[2].c", 23)
        assert sample[r"sec\ond"]()[2]["c"] == 36  # type: ignore
        assert other_sample[r"sec\ond"]()[2]["c"] == 23  # type: ignore

        other_sample = sample.deep_set(r"fir\.st.data[2].new[1].p", [1, 2, 3])
        assert "new" not in sample["fir.st"]()["data"][2]  # type: ignore
        assert "new" in other_sample["fir.st"]()["data"][2]  # type: ignore
        assert len(other_sample["fir.st"]()["data"][2]["new"]) == 2  # type: ignore
        assert other_sample["fir.st"]()["data"][2]["new"][1]["p"] == [  # type: ignore
            1,
            2,
            3,
        ]
        assert other_sample["fir.st"]()["data"][2]["new"][0] is None  # type: ignore

        other_sample = sample.deep_set(r"fir\.st", [1, 2, 3])
        assert not isinstance(sample["fir.st"](), list)
        assert other_sample["fir.st"]() == [1, 2, 3]

    def test_deep_get(self):
        import pipelime.items as pli

        json_data = {"data": ["a", "b", {"c": 42}], "meta": 3.14}
        data = {
            "fir.st": pli.JsonMetadataItem(json_data),
            r"sec\ond": pli.JsonMetadataItem(["a", "b", {"c": 36}]),
        }
        sample = pls.Sample(data)

        assert sample.deep_get(r"fir\.st.data.2.c") == 42
        assert sample.deep_get(r"fir\.st") == json_data
        assert sample.deep_get(r"sec\ond[2].c") == 36
        assert sample.deep_get("not.there", "default") == "default"
        assert sample.deep_get("notthere", "default") == "default"

    def test_match(self):
        sample, data = self._mixed_sample()
        assert sample.match(f"`c.foo` == '{data['c']()['foo']}'")

    def test_change_key_invalid(self):
        sample, _ = self._np_sample()
        other_sample = sample.change_key("__", "--", False)
        assert other_sample is sample

    @pytest.mark.parametrize("delete_old_key", [True, False])
    def test_change_key(self, delete_old_key):
        sample, data = self._np_sample()
        original_key = next(iter(data))
        new_key = "new_key"
        other_sample = sample.change_key(
            original_key, new_key, delete_old_key=delete_old_key
        )

        assert all(k in sample for k in data)
        assert new_key not in sample
        assert all(k in other_sample for k in data if k != original_key)
        assert new_key in other_sample
        assert (
            (original_key not in other_sample)
            if delete_old_key
            else (original_key in other_sample)
        )

    def test_duplicate_key(self):
        sample, data = self._np_sample()
        original_key = next(iter(data))
        new_key = "new_key"
        other_sample = sample.duplicate_key(original_key, new_key)

        assert all(k in sample for k in data)
        assert new_key not in sample
        assert all(k in other_sample for k in data)
        assert new_key in other_sample

    def test_rename_key(self):
        sample, data = self._np_sample()
        original_key = next(iter(data))
        new_key = "new_key"
        other_sample = sample.rename_key(original_key, new_key)

        assert all(k in sample for k in data)
        assert new_key not in sample
        assert all(k in other_sample for k in data if k != original_key)
        assert new_key in other_sample
        assert original_key not in other_sample

    def test_remove_keys(self):
        sample, data = self._np_sample()
        keys_to_remove = [list(data.keys())[0], list(data.keys())[-1]]
        other_sample = sample.remove_keys(*keys_to_remove)

        assert all(k in sample for k in data)
        assert all(
            (k in other_sample)
            if (k not in keys_to_remove)
            else (k not in other_sample)
            for k in data
        )

    def test_extract_keys(self):
        sample, data = self._np_sample()
        keys_to_extract = [list(data.keys())[0], list(data.keys())[-1]]
        other_sample = sample.extract_keys(*keys_to_extract)

        assert all(k in sample for k in data)
        assert all(
            (k in other_sample) if (k in keys_to_extract) else (k not in other_sample)
            for k in data
        )

    def _merge_test(self, fn):
        sample_1, data_1 = self._np_sample()
        sample_2, data_2 = self._mixed_sample()

        other_sample = fn(sample_1, sample_2)

        only_1 = {k for k in data_1.keys() if k not in data_2.keys()}
        only_2 = {k for k in data_2.keys() if k not in data_1.keys()}
        and_1_2 = data_1.keys() & data_2.keys()
        sum_1_2 = data_1.keys() | data_2.keys()

        assert all(k in sample_1 for k in data_1)
        assert all(k not in sample_1 for k in only_2)
        assert all(k in sample_2 for k in data_2)
        assert all(k not in sample_2 for k in only_1)

        assert all(k in other_sample for k in sum_1_2)

        assert all(
            v is sample_2[k] if k in data_2 else v is sample_1[k]
            for k, v in other_sample.items()
        )
        assert all(other_sample[k] is not sample_1[k] for k in and_1_2)

    def test_merge(self):
        self._merge_test(lambda x1, x2: x1.merge(x2))

    def test_update(self):
        self._merge_test(lambda x1, x2: x1.update(x2))



import time

from pipelime.piper.progress.estimator.base import Estimator


class NaiveEstimator(Estimator):
    """A naive estimator that simply estimates the speed and ETA based on the current
    progress and the biased EWA time between steps"""

    def __init__(self, alpha: float = 0.9):
        super().__init__()
        self.alpha = alpha

    def reset(self, total: int) -> None:
        self._avg_dt = -1
        self._progress = 0
        self._total = total
        self._start_time = time.time()
        self._last = self._start_time

    def tick(self, advance: int = 1) -> None:
        now = time.time()
        if advance > 0:
            new_delta = (now - self._last) / advance
            if self._avg_dt < 0:
                self._avg_dt = new_delta
            else:
                self._avg_dt = self._avg_dt * self.alpha + new_delta * (1 - self.alpha)
        self._progress += advance
        self._last = now

    @property
    def start_time(self) -> float:
        return self._start_time

    @property
    def elapsed(self) -> float:
        return self._last - self._start_time

    @property
    def speed(self) -> float:
        return 1 / self._avg_dt if self._avg_dt != 0 else float("inf")

    @property
    def eta(self) -> float:
        if self._avg_dt < 0:
            return -1
        return self._avg_dt * (self._total - self._progress)



import typing as t

import pydantic as pyd

import pipelime.sequences as pls


@pls.source_sequence
class SequenceFromCallable(pls.SamplesSequence, title="from_callable"):
    """A SamplesSequence calling a user-defined generator to get the samples."""

    generator_fn: t.Callable[[int], pls.Sample] = pyd.Field(
        ..., description="A callable returning a sample for a given index."
    )
    length: t.Union[int, t.Callable[[], int]] = pyd.Field(
        ...,
        description=(
            "The length of the sequence, either an integer "
            "or a callable returning an integer."
        ),
    )

    def size(self) -> int:
        return self.length if isinstance(self.length, int) else self.length()

    def get_sample(self, idx: int) -> pls.Sample:
        return self.generator_fn(idx)



from pipelime.stages import SampleStage
from pipelime.sequences import Sample


class MyStage(SampleStage):
    source_key: str
    target_key: str

    def __call__(self, x: Sample) -> Sample:
        from pipelime.items import NumpyItem, ImageItem

        if self.source_key in x:
            item = x[self.source_key]

            # we want numpy items, but not images, though they are ndarrays as well
            if isinstance(item, NumpyItem) and not isinstance(item, ImageItem):
                x = x.set_value_as(
                    self.target_key,
                    self.source_key,
                    float(item()) * 2.5,  # type: ignore
                )
        return x



from pipelime.choixe.visitors.decoder import decode
from pipelime.choixe.visitors.processor import process
from pipelime.choixe.visitors.unparser import unparse
from pipelime.choixe.visitors.walker import walk
from pipelime.choixe.visitors.inspector import inspect, Inspection



from pydantic import Field

import pipelime.sequences as pls
import pipelime.stages as plst


class Invert(plst.SampleStage, title="invert"):
    """Invert the colors of an image."""

    key: str = Field("image", description="The key of the image to invert.")

    def __call__(self, x: pls.Sample) -> pls.Sample:
        return x.set_value(self.key, 255 - x[self.key]())  # type: ignore



from __future__ import annotations
import typing as t
from pathlib import Path
from abc import ABC, abstractmethod
import typer

from pipelime.cli.subcommands import SubCommands as subc

from pipelime.cli.utils import (
    PipelimeSymbolsHelper,
    print_command_op_stage_info,
    print_commands_ops_stages_list,
)


class CLISpecialChars:
    @staticmethod
    def assignment():
        return ("=",)

    @staticmethod
    def config():
        # MUST BE SORTED FROM THE LONGEST TO THE SHORTEST
        return ("++", "+")

    @staticmethod
    def context():
        # MUST BE SORTED FROM THE LONGEST TO THE SHORTEST
        return ("@@", "@")

    @staticmethod
    def ctx_start():
        return ("//",)


class CLIParserState(ABC):
    def __init__(
        self,
        cfg_opts: t.Dict[str, t.Any] = {},
        ctx_opts: t.Dict[str, t.Any] = {},
        ctx_started: bool = False,
    ):
        self.cfg_opts = cfg_opts
        self.ctx_opts = ctx_opts
        self.ctx_started = ctx_started

    @abstractmethod
    def process_token(self, token: str) -> CLIParserState:
        pass

    @abstractmethod
    def close(self):
        pass


class CLIParserHoldState(CLIParserState):
    def process_token(self, token: str) -> CLIParserState:
        from pipelime.cli.pretty_print import print_error

        if token in CLISpecialChars.ctx_start():
            return CLIParserHoldState(self.cfg_opts, self.ctx_opts, True)
        if not self.ctx_started and token.startswith(CLISpecialChars.config()):
            opt, val = self._process_key_arg(token)
            cli_state = CLIParserExpectingCfgValue(
                opt, self.cfg_opts, self.ctx_opts, self.ctx_started
            )
            if val is not None:
                return cli_state.process_token(val)
            return cli_state
        if (
            self.ctx_started and token.startswith(CLISpecialChars.config())
        ) or token.startswith(CLISpecialChars.context()):
            opt, val = self._process_key_arg(token)
            cli_state = CLIParserExpectingCtxValue(
                opt, self.cfg_opts, self.ctx_opts, self.ctx_started
            )
            if val is not None:
                return cli_state.process_token(val)
            return cli_state

        print_error(f"Unexpected token: `{token}`")
        raise typer.Exit(1)

    def close(self):
        return

    def _process_key_arg(self, token: str):
        from pipelime.cli.pretty_print import print_error, print_warning

        opt, val = token, None
        for char in CLISpecialChars.config() + CLISpecialChars.context():
            if token.startswith(char):
                opt = token[len(char) :]  # noqa: E203
                break

        for char in CLISpecialChars.assignment():
            if char in opt:
                opt, _, val = opt.partition(char)
                break

        if opt.endswith(".") or ".." in opt or ".[" in opt:
            print_error(f"Invalid key path: `{opt}`")
            print_warning(
                "Remember: Bash and other shells want the choixe's dollar sign (`$`) "
                "escaped! Try with single quotes or backslash."
            )
            print_warning("For example:")
            print_warning(
                "bash/zsh: +operations.map.$model => '+operations.map.$model'"
            )
            print_warning(r"zsh: +operations.map.$model => +operations.map.\$model")
            raise typer.Exit(1)

        return opt, val


class CLIParserExpectingValue(CLIParserState):
    def __init__(
        self,
        key_name: str,
        cfg_opts: t.Dict[str, t.Any] = {},
        ctx_opts: t.Dict[str, t.Any] = {},
        ctx_started: bool = False,
    ):
        super().__init__(cfg_opts, ctx_opts, ctx_started)
        self.key_name = key_name

    @abstractmethod
    def target_cfg(self) -> t.Dict[str, t.Any]:
        pass

    def process_token(self, token: str) -> CLIParserState:
        from pipelime.choixe.utils.common import deep_set_

        if token in CLISpecialChars.ctx_start() or token.startswith(
            CLISpecialChars.config() + CLISpecialChars.context()
        ):
            self._set_boolean_flag()
            cli_state = CLIParserHoldState(
                self.cfg_opts, self.ctx_opts, self.ctx_started
            )
            return cli_state.process_token(token)

        deep_set_(
            self.target_cfg(),
            key_path=self.key_name,
            value=self._convert_val(token),
            append=True,
        )
        return CLIParserHoldState(self.cfg_opts, self.ctx_opts, self.ctx_started)

    def close(self):
        self._set_boolean_flag()

    def _set_boolean_flag(self):
        from pipelime.choixe.utils.common import deep_set_

        deep_set_(
            self.target_cfg(),
            key_path=self.key_name,
            value=True,
            append=True,
        )

    def _convert_val(self, val: str):
        if val.lower() == "true":
            return True
        if val.lower() == "false":
            return False
        if val.lower() in ("none", "null", "nul"):
            return None
        try:
            num = int(val)
            return num
        except ValueError:
            pass
        try:
            num = float(val)
            return num
        except ValueError:
            pass
        return val


class CLIParserExpectingCfgValue(CLIParserExpectingValue):
    def target_cfg(self) -> t.Dict[str, t.Any]:
        return self.cfg_opts


class CLIParserExpectingCtxValue(CLIParserExpectingValue):
    def target_cfg(self) -> t.Dict[str, t.Any]:
        return self.ctx_opts


def _complete_yaml(incomplete: str):
    for v in Path(".").glob(f"{incomplete}*.yaml"):
        yield str(v)


def _print_dict(name, data):
    import json
    from pipelime.cli.pretty_print import print_info

    print_info(f"\n{name}:")
    print_info(json.dumps(data, indent=2))


def _update_dispatch(original_value, item):
    if isinstance(item, t.Mapping):
        if isinstance(original_value, t.MutableMapping):
            _dict_update(original_value, item)
            return True
    elif isinstance(item, t.Sequence):
        if isinstance(original_value, t.MutableSequence):
            _list_update(original_value, item)
            return True
    return False


def _list_update(to_be_updated: t.MutableSequence, data: t.Sequence):
    to_be_updated.extend([None] * (len(data) - len(to_be_updated)))  # this is safe
    if data[-1] is None:
        del to_be_updated[len(data) - 1 :]  # noqa: E203
    for idx, item in enumerate(data):
        if item is not None:
            original_value = to_be_updated[idx]
            if _update_dispatch(original_value, item):
                continue
            to_be_updated[idx] = item


def _dict_update(to_be_updated: t.MutableMapping, data: t.Mapping):
    for k, v in data.items():
        if k in to_be_updated:
            original_value = to_be_updated[k]
            if _update_dispatch(original_value, v):
                continue
        to_be_updated[k] = v


def _process_cfg_or_die(
    cfg,
    ctx,
    run_all: t.Optional[bool],
    output: t.Optional[Path],
    exit_on_error: bool = True,
):
    from pipelime.cli.pretty_print import print_error
    from pipelime.choixe.visitors.processor import ChoixeProcessingError

    try:
        effective_configs = (
            [cfg.process(ctx)]
            if run_all is not None and not run_all
            else cfg.process_all(ctx)
        )
    except ChoixeProcessingError as e:
        if exit_on_error:
            print_error(f"Invalid configuration! {e}")
            raise typer.Exit(1)
        raise e

    if output is not None:
        zero_fill = len(str(len(effective_configs) - 1))
        for idx, cfg in enumerate(effective_configs):
            filepath = (
                output.with_name(f"{output.name}_{str(idx).zfill(zero_fill)}")
                if len(effective_configs) > 1
                else output
            )
            cfg.save_to(filepath)
    return effective_configs


app = typer.Typer(pretty_exceptions_enable=False)


def version_callback(value: bool):
    from pipelime import __version__

    if value:
        print(__version__)
        raise typer.Exit()


@app.command(
    add_help_option=False,
    no_args_is_help=True,
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
)
def pl_main(  # noqa: C901
    ctx: typer.Context,
    config: t.Optional[Path] = typer.Option(
        None,
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
        help=(
            "A yaml/json file with some or all the arguments "
            "required by the command.\n\n"
            "`++opt` or `+opt` command line options update and override them.\n\n "
        ),
        autocompletion=_complete_yaml,
    ),
    context: t.Optional[Path] = typer.Option(
        None,
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
        help=(
            "A yaml/json file with some or all the context parameters.\n\n"
            "`@@opt` or `@opt` command line options update and override them.\n\n"
            "After a `//` token, `++opt` and `+opt` are accepted as well.\n\n "
        ),
        autocompletion=_complete_yaml,
    ),
    extra_modules: t.List[str] = typer.Option(
        [], "--module", "-m", help="Additional modules to import."
    ),
    run_all: t.Optional[bool] = typer.Option(
        None,
        help=(
            "In case of multiple configurations, run them all, "
            "otherwise, run only the first one. If not specified, user will be "
            "notified if multiple configurations are found."
        ),
    ),
    output: t.Optional[Path] = typer.Option(
        None,
        "--output",
        "-o",
        writable=True,
        resolve_path=True,
        help="Save final processed configuration to json/yaml.",
    ),
    verbose: bool = typer.Option(False, "--verbose", "-v", help="Verbose output."),
    dry_run: bool = typer.Option(False, "--dry-run", "-d", help="Dry run."),
    command: str = typer.Argument(
        "",
        show_default=False,
        help=(
            (
                "A pipelime command, ie, a `command-name`, "
                "a `package.module.ClassName` class path or "
                "a `path/to/module.py:ClassName` uri (use with care).\n\n"
            )
            + subc.get_help()
        ),
        autocompletion=PipelimeSymbolsHelper.complete_name(
            is_cmd=True,
            is_seq_op=False,
            is_stage=False,
            additional_names=subc.get_autocompletions(),
        ),
    ),
    command_args: t.Optional[t.List[str]] = typer.Argument(
        None,
        help=(
            "\b\nPipelime command arguments:\n"
            "- `++opt` and `+opt` are command parameters\n"
            "- `@@opt` and `@opt` are context parameters\n"
            "- after `//` `++opt`, `+opt`, `@@opt`and `@opt` "
            "are always context parameters."
        ),
    ),
    help: bool = typer.Option(
        False, "--help", "-h", show_default=False, help="Show this message and exit."
    ),
    version: t.Optional[bool] = typer.Option(
        None,
        "--version",
        show_default=False,
        help="Show version and exit.",
        callback=version_callback,
        is_eager=True,
    ),
):
    """
    Pipelime Command Line Interface. Examples:

    `pipelime list` prints a list of the available commands, sequence operators
    and stages.

    `pipelime help <cmd-op-stg>` prints informations on a specific command, sequence
    operator or stage.

    `pipelime <command> [<args>]` runs a pipelime command.

    NB: command (++opt) and context (@@opt) arguments with no value are treated as
    TRUE boolean values. Use `false` or `true` to explicitly set a boolean
    and `none`/`null`/`nul` to enforce `None`.
    """

    PipelimeSymbolsHelper.set_extra_modules(extra_modules)

    if command_args is None:
        command_args = []

    if (
        help
        or command in subc.HELP[0]
        or any([h in command_args for h in subc.HELP[0]])
    ):
        try:
            if command and command not in subc.HELP[0]:
                print_command_op_stage_info(command)
            elif command_args:
                print_command_op_stage_info(command_args[0])
            else:
                print(ctx.get_help())
        except ValueError:
            pass
    elif command in subc.WIZARD[0] or any([w in command_args for w in subc.WIZARD[0]]):
        from pipelime.cli.wizard import Wizard

        if command and command not in subc.WIZARD[0]:
            Wizard.model_cfg_wizard(command)
        Wizard.model_cfg_wizard(command_args[0])
    elif command in subc.LIST[0]:
        print_commands_ops_stages_list(
            verbose, show_cmds=True, show_ops=True, show_stages=True
        )
    elif command in subc.LIST_CMDS[0]:
        print_commands_ops_stages_list(
            verbose, show_cmds=True, show_ops=False, show_stages=False
        )
    elif command in subc.LIST_OPS[0]:
        print_commands_ops_stages_list(
            verbose, show_cmds=False, show_ops=True, show_stages=False
        )
    elif command in subc.LIST_STGS[0]:
        print_commands_ops_stages_list(
            verbose, show_cmds=False, show_ops=False, show_stages=True
        )
    elif command:
        from pipelime.choixe import XConfig
        import pipelime.choixe.utils.io as choixe_io
        from pipelime.cli.pretty_print import print_error, print_warning, print_info

        if verbose:
            print_info(
                "No configuration file"
                if config is None
                else f"Configuration file: {config}"
            )
            print_info(
                "No context file" if context is None else f"Context file: {config}"
            )
            print_info(
                f"Other command and context arguments: {command_args}"
                if command_args
                else "No other command or context arguments"
            )

        base_cfg = {} if config is None else choixe_io.load(config)
        base_ctx = {} if context is None else choixe_io.load(context)

        # process extra args
        cli_state = CLIParserHoldState()
        for token in command_args:
            cli_state = cli_state.process_token(token)
        cli_state.close()
        cmdline_cfg, cmdline_ctx = cli_state.cfg_opts, cli_state.ctx_opts

        if verbose:
            _print_dict("Loaded configuration file", base_cfg)
            _print_dict("Loaded context file", base_ctx)
            _print_dict("Configuration options from command line", cmdline_cfg)
            _print_dict("Context options from command line", cmdline_ctx)

        _dict_update(base_cfg, cmdline_cfg)
        _dict_update(base_ctx, cmdline_ctx)

        if verbose:
            _print_dict("Merged configuration", base_cfg)
            _print_dict("Merged context", base_ctx)

        base_cfg = XConfig(
            data=base_cfg, cwd=Path.cwd() if config is None else config.parent
        )
        base_ctx = XConfig(
            data=base_ctx, cwd=Path.cwd() if context is None else context.parent
        )

        if command in subc.AUDIT[0]:
            from dataclasses import fields
            from pipelime.choixe.visitors.processor import ChoixeProcessingError

            print_info("\n📄 CONFIGURATION AUDIT\n")
            inspect_info = base_cfg.inspect()
            for field in fields(inspect_info):
                value = getattr(inspect_info, field.name)
                print_info(f"🔍 {field.name}:")
                if value or isinstance(value, bool):
                    print_info(value, pretty=True)

            print_info("\n📄 CONTEXT AUDIT\n")
            print_info(base_ctx.to_dict(), pretty=True)
            print_info("")

            try:
                effective_configs = _process_cfg_or_die(
                    base_cfg, base_ctx, run_all, output, False
                )
            except ChoixeProcessingError as e:
                from rich.prompt import Prompt, Confirm
                from pipelime.cli.wizard import Wizard

                print_warning("Some variables are not defined in the context.")
                print_error(f"Invalid configuration! {e}")
                raise typer.Exit(1)

                ### SKIP FOR NOW
                # if not Confirm.ask(
                #     "Do you want to create a new context?", default=True
                # ):
                #     print_error(f"Invalid configuration! {e}")
                #     raise typer.Exit(1)
            #
            # print_info("\n📝 Please enter a value for each variable")
            # new_ctx = Wizard.context_wizard(inspect_info.variables, base_ctx)
            #
            # print_info("Processing configuration and context...", end="")
            # effective_configs = _process_cfg_or_die(
            #     base_cfg, new_ctx, run_all, output
            # )
            # print_info(" OK")
            #
            # outfile = Prompt.ask("\n💾 Write to (leave empty to skip)")
            # if outfile:
            #     new_ctx.save_to(Path(outfile).with_suffix(".yaml"))

            pls = "s" if len(effective_configs) != 1 else ""
            print_info(
                "🎉 Configuration successfully processed "
                f"({len(effective_configs)} variant{pls})."
            )
            raise typer.Exit(0)
        else:
            effective_configs = _process_cfg_or_die(base_cfg, base_ctx, run_all, output)

            if verbose:
                pls = "s" if len(effective_configs) != 1 else ""
                print_info(f"Found {len(effective_configs)} configuration{pls}")

            for cfg in effective_configs:
                if not cfg.inspect().processed:
                    print_error("The configuration has not been fully processed.")
                    if verbose:
                        _print_dict("Unprocessed data", cfg)
                    else:
                        print_info(
                            "Run with --verbose to see the final configuration file."
                        )
                    raise typer.Exit(1)

            if len(effective_configs) > 1 and run_all is None:
                if not typer.confirm(
                    f"{len(effective_configs)} configurations found. "
                    "Do you want to run them all?"
                ):
                    effective_configs = effective_configs[:1]

            cfg_size = len(effective_configs)
            for idx, cfg in enumerate(effective_configs):
                if verbose and cfg_size > 1:
                    print_info(f"*** CONFIGURATION {idx}/{cfg_size} ***")
                cfg_dict = cfg.to_dict()
                cmd_name = command

                if cmd_name in subc.EXEC[0]:
                    if len(cfg_dict) == 0:
                        print_error("No command specified.")
                        raise typer.Exit(1)
                    if len(cfg_dict) > 1:
                        print_error("Multiple commands found.")
                        print_warning(
                            "You should use the `run` command to process a dag."
                        )
                        raise typer.Exit(1)
                    cmd_name = next(iter(cfg_dict))
                    cfg_dict = next(iter(cfg_dict.values()))

                run_command(cmd_name, cfg_dict, verbose, dry_run)
    else:
        from pipelime.cli.pretty_print import print_error

        print_error("No command specified.")
        raise typer.Exit(1)


def run_command(command: str, cmd_args: t.Mapping, verbose: bool, dry_run: bool):
    """
    Run a pipelime command.
    """

    import time

    from pipelime.cli.pretty_print import (
        print_info,
        print_command_outputs,
    )
    from pipelime.piper import PipelimeCommand
    from pipelime.cli.utils import PipelimeSymbolsHelper, time_to_str

    cmd_cls = PipelimeSymbolsHelper.get_command(command)
    if cmd_cls is None or not issubclass(cmd_cls[1], PipelimeCommand):
        PipelimeSymbolsHelper.show_error_and_help(
            command, should_be_cmd=True, should_be_op=False, should_be_stage=False
        )
        raise typer.Exit(1)
    cmd_cls = cmd_cls[1]

    if verbose:
        print_info(f"\nCreating command `{command}` with options:")
        print_info(cmd_args, pretty=True)

    cmd_obj = cmd_cls(**cmd_args)

    if verbose:
        print_info(f"\nCreated command `{command}`:")
        print_info(cmd_obj.dict(), pretty=True)

    if verbose:
        print_info(f"\nRunning `{command}`...")

    start_time = time.perf_counter_ns()
    if not dry_run:
        cmd_obj()
    end_time = time.perf_counter_ns()
    print_info("\nCommand executed in " + time_to_str(end_time - start_time))

    print_info(f"\n`{command}` outputs:")
    print_command_outputs(cmd_obj)


if __name__ == "__main__":
    app()



import pickle
import typing as t

from pipelime.items import Item


class PickleItem(Item[t.Any]):
    @classmethod
    def file_extensions(cls) -> t.Sequence[str]:
        return ('.pkl', '.pickle')

    @classmethod
    def decode(cls, fp: t.BinaryIO) -> t.Any:
        return pickle.load(fp)

    @classmethod
    def encode(cls, value: t.Any, fp: t.BinaryIO):
        pickle.dump(value, fp)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, Sequence, Tuple, Union

# from dataclasses import dataclass
from pydantic.dataclasses import dataclass


class NodeVisitor:  # pragma: no cover
    """Base class for all Choixe visitors. A visitor can be seen as a special recursive
    function that transforms a `Node` (i.e. an element of a syntax tree) into something
    else, depending on the implementation.

    A `NodeVisitor` implements a variation of the Visitor design pattern on a generic
    `Node` object, i.e. an element in the Choixe AST. In particular, visiting methods
    can return values, providing an easier way to use visitors to produce an output.

    This basic node visitor simply forwards the inputs to the outputs, essentially
    performing no operation on the AST.

    All the methods of `NodeVisitor` are called `visit_<node_type>`, they take as input
    the corresponding node object, and they can return anything.
    """

    def _ignore(self, node: Node) -> Any:
        return node

    def visit_dict(self, node: DictNode) -> Any:
        return self._ignore(node)

    def visit_list(self, node: ListNode) -> Any:
        return self._ignore(node)

    def visit_literal(self, node: LiteralNode) -> Any:
        return self._ignore(node)

    def visit_dict_bundle(self, node: DictBundleNode) -> Any:
        return self._ignore(node)

    def visit_str_bundle(self, node: StrBundleNode) -> Any:
        return self._ignore(node)

    def visit_var(self, node: VarNode) -> Any:
        return self._ignore(node)

    def visit_import(self, node: ImportNode) -> Any:
        return self._ignore(node)

    def visit_sweep(self, node: SweepNode) -> Any:
        return self._ignore(node)

    def visit_symbol(self, node: SymbolNode) -> Any:
        return self._ignore(node)

    def visit_instance(self, node: InstanceNode) -> Any:
        return self._ignore(node)

    def visit_model(self, node: ModelNode) -> Any:
        return self._ignore(node)

    def visit_for(self, node: ForNode) -> Any:
        return self._ignore(node)

    def visit_switch(self, node: SwitchNode) -> Any:
        return self._ignore(node)

    def visit_index(self, node: IndexNode) -> Any:
        return self._ignore(node)

    def visit_item(self, node: ItemNode) -> Any:
        return self._ignore(node)

    def visit_uuid(self, node: UuidNode) -> Any:
        return self._ignore(node)

    def visit_date(self, node: DateNode) -> Any:
        return self._ignore(node)

    def visit_cmd(self, node: CmdNode) -> Any:
        return self._ignore(node)

    def visit_tmp_dir(self, node: TmpDirNode) -> Any:
        return self._ignore(node)


@dataclass
class Node(ABC):
    """A generic element of the Choixe AST, all nodes must implement this interface."""

    @abstractmethod
    def accept(self, visitor: NodeVisitor) -> Any:  # pragma: no cover
        """Accepts an incoming visitor. This method should simply call the respective
        `visit_<node_type>` method of the visiting object, pass `self` as argument
        and forward the result.

        Args:
            visitor (NodeVisitor): The visiting object.

        Returns:
            Any: The visitor result.
        """
        pass


@dataclass
class HashNode(Node):
    """A `HashNode` is a special type of node that is hashable. It represents an
    immutable structure and can generally be used as a dictionary key."""

    def __hash__(self) -> int:
        return hash(tuple(self.__dict__.values()))

    def __eq__(self, __o: object) -> bool:
        return self.__class__ == __o.__class__ and hash(self) == hash(__o)


@dataclass
class DictNode(Node):
    """AST node for Mapping-like structures. Contains a mapping from `HashNode` to `Node`."""

    nodes: Dict[HashNode, Node]

    def accept(self, visitor: NodeVisitor) -> Any:
        return visitor.visit_dict(self)


@dataclass(init=False)
class ListNode(Node):
    """AST node for List-like structures. Contains a list of `Node` objects."""

    nodes: Sequence[Node]

    def __init__(self, *nodes: Node) -> None:
        self.nodes = nodes

    def accept(self, visitor: NodeVisitor) -> Any:
        return visitor.visit_list(self)


@dataclass(eq=False)
class LiteralNode(HashNode):
    """An `LiteralNode` contains a single generic hashable python object, like a built-in
    int, float or str."""

    data: Any

    def accept(self, visitor: NodeVisitor) -> Any:
        return visitor.visit_literal(self)


@dataclass(init=False, eq=False)
class DictBundleNode(Node):
    """An `DictBundleNode` represents a dictionary union of a nodes collection."""

    nodes: Sequence[Node]

    def __init__(self, *nodes: Node) -> None:
        self.nodes = nodes

    def accept(self, visitor: NodeVisitor) -> Any:
        return visitor.visit_dict_bundle(self)


@dataclass(init=False, eq=False)
class StrBundleNode(HashNode):
    """A `StrBundleNode` represents a concatenation of a sequence of strings."""

    nodes: Sequence[HashNode]

    def __init__(self, *nodes: HashNode) -> None:
        self.nodes = nodes

    def accept(self, visitor: NodeVisitor) -> Any:
        return visitor.visit_str_bundle(self)


@dataclass(eq=False)
class VarNode(HashNode):
    """A `VarNode` represents a Choixe variable. It has an id and a default value."""

    identifier: HashNode
    default: Optional[HashNode] = None
    env: Optional[HashNode] = None
    help: Optional[LiteralNode] = None

    def accept(self, visitor: NodeVisitor) -> Any:
        return visitor.visit_var(self)


@dataclass
class ImportNode(Node):
    """An `ImportNode` represents a Choixe import directive from a filesystem path."""

    path: HashNode

    def accept(self, visitor: NodeVisitor) -> Any:
        return visitor.visit_import(self)


@dataclass(init=False, eq=False)
class SweepNode(HashNode):
    """A `SweepNode` represents a Choixe sweep directive from multiple branching options."""

    cases: Sequence[Node]

    def __init__(self, *cases: Node) -> None:
        self.cases = cases

    def accept(self, visitor: NodeVisitor) -> Any:
        return visitor.visit_sweep(self)


@dataclass
class SymbolNode(Node):
    """A `SymbolNode` represents a Choixe instance block to import a generic python object."""

    symbol: HashNode

    def accept(self, visitor: NodeVisitor) -> Any:
        return visitor.visit_symbol(self)


@dataclass
class InstanceNode(Node):
    """An `InstanceNode` represents a Choixe instance block to get the result of a
    generic python callable object."""

    symbol: HashNode
    args: Union[DictNode, DictBundleNode]

    def accept(self, visitor: NodeVisitor) -> Any:
        return visitor.visit_instance(self)


@dataclass
class ModelNode(InstanceNode):
    def accept(self, visitor: NodeVisitor) -> Any:
        return visitor.visit_model(self)


@dataclass
class ForNode(Node):
    """A `ForNode` represents a Choixe for loop that iterates over a collection picked
    from the context. A for loop also has an string identifier, that mast be a valid
    python id."""

    iterable: LiteralNode
    body: Node
    identifier: Optional[LiteralNode] = None

    def accept(self, visitor: NodeVisitor) -> Any:
        return visitor.visit_for(self)


@dataclass
class SwitchNode(Node):

    value: HashNode
    cases: Sequence[Tuple[Node, Node]]
    default: Optional[Node] = None

    def accept(self, visitor: NodeVisitor) -> Any:
        return visitor.visit_switch(self)


@dataclass(eq=False)
class IndexNode(HashNode):
    """An `IndexNode` represents the index of the current iteration of a for loop."""

    identifier: Optional[HashNode] = None

    def accept(self, visitor: NodeVisitor) -> Any:
        return visitor.visit_index(self)


@dataclass(eq=False)
class ItemNode(HashNode):
    """An `ItemNode` represents the item of the current iteration of a for loop."""

    identifier: Optional[HashNode] = None

    def accept(self, visitor: NodeVisitor) -> Any:
        return visitor.visit_item(self)


@dataclass(eq=False)
class UuidNode(HashNode):
    """An `UuidNode` represents a randomly generated uuid."""

    def accept(self, visitor: NodeVisitor) -> Any:
        return visitor.visit_uuid(self)


@dataclass(eq=False)
class DateNode(HashNode):
    """A `DateNode` represents the current datetime with an optional custom format."""

    format: Optional[HashNode] = None

    def accept(self, visitor: NodeVisitor) -> Any:
        return visitor.visit_date(self)


@dataclass(eq=False)
class CmdNode(HashNode):
    """A `CmdNode` represents the calling of a system command."""

    command: HashNode

    def accept(self, visitor: NodeVisitor) -> Any:
        return visitor.visit_cmd(self)


@dataclass(eq=False)
class TmpDirNode(HashNode):
    """A `TmpDirNode` represents the creation of a temporary directory."""

    name: Optional[HashNode] = None

    def accept(self, visitor: NodeVisitor) -> Any:
        return visitor.visit_tmp_dir(self)



import uuid
from enum import Enum
from pathlib import Path
from typing import Optional

import typer
from typer import Option

piper = typer.Typer()


dag_help = "The DAG configuration file"
params_help = "The parameters configuration file"
draw_backend_help = "The backend to use for drawing the graph"
draw_open_help = "Open the generated image"
draw_output_help = "The output file"
compile_out_help = "The output file"


@piper.command()
def compile(
    dag_file: Path = Option(..., "-d", "--dag", help=dag_help),
    params_file: Path = Option(..., "-p", "--params", help=params_help),
    output_file: Optional[Path] = Option(None, "-o", "--output", help=compile_out_help),
):
    import rich

    from pipelime.choixe import XConfig
    from pipelime.piper.parsers.factory import DAGParserFactory

    dag = DAGParserFactory.get_parser().parse_file(dag_file, params_file)
    compiled = XConfig(dag.purged_dict())

    if output_file is not None:
        compiled.save_to(output_file)
    else:
        rich.print(compiled)


@piper.command()
def run(
    dag_file: Path = Option(..., "-d", "--dag", help=dag_help),
    params_file: Path = Option(..., "-p", "--params", help=params_help),
    watch: bool = Option(False, "-w", "--watch", is_flag=True, help="Watch the DAG"),
    token: str = Option(
        uuid.uuid1().hex, "-t", "--token", help="The piper execution token"
    ),
):
    from pipelime.piper.executors.factory import NodesGraphExecutorFactory
    from pipelime.piper.graph import DAGNodesGraph
    from pipelime.piper.parsers.factory import DAGParserFactory

    dag = DAGParserFactory.get_parser().parse_file(dag_file, params_file)
    graph = DAGNodesGraph.build_nodes_graph(dag)
    executor = NodesGraphExecutorFactory.get_executor(watch=watch)
    executor.exec(graph, token=token)


class DrawBackendChoice(Enum):
    GRAPHVIZ = "graphviz"
    MERMAID = "mermaid"


@piper.command()
def draw(
    dag_file: Path = Option(..., "-d", "--dag", help=dag_help),
    params_file: Optional[Path] = Option(None, "-p", "--params", help=params_help),
    output_file: Optional[Path] = Option(None, "-o", "--output", help=draw_output_help),
    draw_backend: DrawBackendChoice = Option(
        "graphviz", "-b", "--backend", help=draw_backend_help
    ),
    open: bool = typer.Option(False, is_flag=True, help=draw_open_help),
):
    import os
    import platform
    import subprocess

    import numpy as np
    import rich
    from PIL import Image

    from pipelime.piper.drawing.factory import NodesGraphDrawerFactory
    from pipelime.piper.graph import DAGNodesGraph
    from pipelime.piper.parsers.factory import DAGParserFactory

    def start_file(filename: str):
        if platform.system() == "Darwin":  # macOS
            subprocess.call(("open", filename))
        elif platform.system() == "Windows":  # Windows
            os.startfile(filename)
        else:  # linux variants #TODO: verify!
            subprocess.call(("xdg-open", filename))

    # DAG Model (abstraction layer to allow several parsing methods)
    dag = DAGParserFactory.get_parser().parse_file(
        cfg_file=dag_file, params_file=params_file
    )

    # Graph
    graph = DAGNodesGraph.build_nodes_graph(dag)

    # Drawn image
    graph_image: Optional[np.ndarray] = None

    # Draw with selected backend
    drawer = NodesGraphDrawerFactory.create(draw_backend.value)

    # Show or Write
    if output_file is not None:
        drawer.export(graph, str(output_file))
        if open:
            start_file(str(output_file))
        else:
            rich.print("graph image saved to:", output_file)
    else:
        graph_image = drawer.draw(graph=graph)
        img = Image.fromarray(graph_image, "RGB")
        img.show("Graph")


if __name__ == "__main__":
    piper()



import time
from threading import Thread

import zmq

from pipelime.piper.progress.listener.receiver.zmq import ZMQProgressReceiver
from pipelime.piper.progress.model import OperationInfo, ProgressUpdate


class TestZMQProgressReceiver:
    N_PACKETS = 5

    def _sending_thread(self, op_info: OperationInfo) -> None:
        context = zmq.Context()
        socket = context.socket(zmq.PUB)
        socket.bind("tcp://*:5556")
        time.sleep(1)
        for i in range(2 * self.N_PACKETS):
            token = "token" if i % 2 == 0 else "token2"
            prog = ProgressUpdate(op_info=op_info, progress=i // 2)
            socket.send_multipart([token.encode(), prog.json().encode()])
            time.sleep(0.1)

    def test_receiver(self):
        op_info = OperationInfo(
            token="token",
            node="node",
            chunk=0,
            message="Message",
            total=self.N_PACKETS,
        )
        receiver = ZMQProgressReceiver("token")

        thread = Thread(target=self._sending_thread, args=[op_info])
        thread.start()

        try:
            packets = []
            t = time.time()
            while len(packets) < self.N_PACKETS and time.time() - t < 10:
                prog = receiver.receive()
                if prog is not None:
                    packets.append(prog)
                    print(prog)

            for i, prog in enumerate(packets):
                assert isinstance(prog, ProgressUpdate)
                assert prog.op_info == op_info
                assert prog.progress == i
        finally:
            thread.join()



from pathlib import Path

import pipelime.choixe as choixe
import pytest
from pipelime.choixe.utils.imports import import_symbol


def test_import_symbol_fn():
    res = import_symbol("numpy.array")
    from numpy import array

    assert res is array


def test_import_symbol_cls():
    res = import_symbol("rich.console.Console")
    from rich.console import Console

    assert res is Console


def test_import_symbol_path():
    file_path = Path(choixe.__file__).parent / "ast" / "nodes.py"
    res = import_symbol(f"{str(file_path)}:SweepNode")
    assert res.__name__ == "SweepNode"


def test_import_symbol_raise():
    # Non existing module
    with pytest.raises(ImportError):
        import_symbol("rich.console.SomethingThatDoesNotExist")

    # Non existing path
    with pytest.raises(ImportError):
        import_symbol("file_that_does_not_exist.py:AAAAAA")

    # Non existing object
    file_path = Path(choixe.__file__).parent / "ast" / "nodes.py"
    with pytest.raises(ImportError):
        import_symbol(f"{str(file_path)}:HolyPinoly")



from pydantic import Field

from pipelime.commands.interfaces import InputDatasetInterface, OutputDatasetInterface
from pipelime.piper import PipelimeCommand, PiperPortType


class RandomSlice(PipelimeCommand, title="randrange"):
    """Extracts a random range from the input dataset."""

    input: InputDatasetInterface = Field(
        ..., description="Input dataset.", piper_port=PiperPortType.INPUT
    )
    output: OutputDatasetInterface = Field(
        ..., description="Output dataset.", piper_port=PiperPortType.OUTPUT
    )
    seed: int = Field(42, description="Random seed.")

    def run(self):
        import random

        seq = self.input.create_reader()

        random.seed(self.seed)
        seq = seq[0 : random.randrange(0, len(seq))]  # noqa: E203
        seq = self.output.append_writer(seq)

        for _ in seq:
            pass



from pathlib import Path

import pytest
from pipelime.choixe.ast.parser import parse
from pipelime.choixe.visitors import Inspection, inspect


class TestInspector:
    @pytest.mark.parametrize(
        ["expr", "expected"],
        [
            [{"a": 10, "b": {"a": 20.0}}, Inspection(processed=True)],
            [
                {
                    "$var(variable.one)": 10,
                    "b": "$var(variable.two, default=10.2)",
                    "c": "$var(variable.three, env=True)",
                },
                Inspection(
                    variables={"variable": {"one": None, "two": 10.2, "three": None}},
                    environ={"variable.three": None},
                ),
            ],
            [
                [
                    "$var(variable.one)",
                    "$var(variable.two, default=10.2)",
                    "$var(variable.three, env=True)",
                ],
                Inspection(
                    variables={"variable": {"one": None, "two": 10.2, "three": None}},
                    environ={"variable.three": None},
                ),
            ],
            [
                "String with $var(variable.one) $var(variable.two, default=10.2) $var(variable.three, env=True)",
                Inspection(
                    variables={"variable": {"one": None, "two": 10.2, "three": None}},
                    environ={"variable.three": None},
                ),
            ],
            [
                {
                    "$directive": "sweep",
                    "$args": [10, "$var(x)", "$var(variable.x)"],
                    "$kwargs": {},
                },
                Inspection(variables={"x": None, "variable": {"x": None}}),
            ],
            [
                "$symbol(my_package.my_symbol)",
                Inspection(symbols={"my_package.my_symbol"}),
            ],
            [
                {"$call": "numpy.array", "$args": {"shape": [4, 3, 2]}},
                Inspection(symbols={"numpy.array"}),
            ],
            [
                {"$model": "path/to/my_file.py:MyModel", "$args": {"shape": [4, 3, 2]}},
                Inspection(symbols={"path/to/my_file.py:MyModel"}),
            ],
            [
                {"$for(var.x.y, x)": {"$index(x)": "$item(x)"}},
                Inspection(variables={"var": {"x": {"y": None}}}),
            ],
            [
                {
                    "$switch(nation)": [
                        {
                            "$case": ["UK", "$var(a_nation)", "Australia"],
                            "$then": "hello",
                        },
                        {
                            "$case": "Italy",
                            "$then": "$var(italian_salute)",
                        },
                        {
                            "$default": "$var(default_salute)",
                        },
                    ]
                },
                Inspection(
                    variables={
                        "nation": None,
                        "a_nation": None,
                        "italian_salute": None,
                        "default_salute": None,
                    }
                ),
            ],
            [
                {
                    "$for(var.my_var, x)": {"$index(x)": "$item(x)"},
                    "$var(var.another_var)": 10,
                    "$for(var.my_var2, y)": {"$index(y)": "$var(var.another_var2)"},
                },
                Inspection(
                    variables={
                        "var": {
                            "my_var": None,
                            "my_var2": None,
                            "another_var": None,
                            "another_var2": None,
                        }
                    }
                ),
            ],
            [
                {
                    "a": "$var(x.sub_var)",
                    "b": "$var(x)",
                },
                Inspection(variables={"x": {"sub_var": None}}),
            ],
            [
                {
                    "foo": "$var(my_var_a)",
                    "$for(my_var_b, x)": {
                        "b_$index(x)": "$item(x.sub_var_a)",
                        "a_$index(x)": "$item(x)",
                    },
                    "$for(my_var_c, y)": {
                        "a_$index(y)": "$item(y.sub_var_a)",
                        "c_$index(y)": "$item(y.sub_var_b.sub_sub_var_a)",
                        "b_$index(y)": "$item(y.sub_var_b)",
                    },
                    "$for(my_var_c, z)": {
                        "b_$index(z)": "$item(z.sub_var_a.sub_sub_var_b)",
                        "c_$index(z)": "$item(z.sub_var_b)",
                        "a_$index(z)": "$item(z.sub_var_a.sub_sub_var_a)",
                    },
                },
                Inspection(
                    variables={
                        "my_var_a": None,
                        "my_var_b": {"sub_var_a": None},
                        "my_var_c": {
                            "sub_var_a": {"sub_sub_var_a": None, "sub_sub_var_b": None},
                            "sub_var_b": {"sub_sub_var_a": None},
                        },
                    }
                ),
            ],
        ],
    )
    def test_inspector(self, expr, expected):
        print(inspect(parse(expr)))
        print(expected)
        assert inspect(parse(expr)) == expected

    def test_relative_import(self, choixe_plain_cfg):
        data = {
            "a": f"$import('{choixe_plain_cfg.name}')",
            "b": {"c": f"$import('{choixe_plain_cfg.name}')"},
        }
        expected = Inspection(imports={choixe_plain_cfg})
        assert inspect(parse(data), cwd=choixe_plain_cfg.parent) == expected

    def test_import_not_found(self):
        data = {"a": "$import('nonexisting.yml')"}
        expected = Inspection(imports={Path("nonexisting.yml").absolute()})
        with pytest.warns(UserWarning):
            assert inspect(parse(data)) == expected



from typing import Any

import pytest

import pipelime.choixe.ast.nodes as ast
from pipelime.choixe.ast.parser import ChoixeParsingError, parse


class TestParse:
    def test_parse_dict(self):
        expr = {"a": 10, "b": {"c": 10.0, "d": "hello"}}
        expected = ast.DictNode(
            nodes={
                ast.LiteralNode(data="a"): ast.LiteralNode(data=10),
                ast.LiteralNode(data="b"): ast.DictNode(
                    nodes={
                        ast.LiteralNode(data="c"): ast.LiteralNode(data=10.0),
                        ast.LiteralNode(data="d"): ast.LiteralNode(data="hello"),
                    }
                ),
            }
        )
        assert parse(expr) == expected

    def test_parse_list(self):
        expr = [1, 2, 3, ("foo", "bar", [10.0, 10])]
        expected = ast.ListNode(
            ast.LiteralNode(data=1),
            ast.LiteralNode(data=2),
            ast.LiteralNode(data=3),
            ast.ListNode(
                ast.LiteralNode(data="foo"),
                ast.LiteralNode(data="bar"),
                ast.ListNode(ast.LiteralNode(data=10.0), ast.LiteralNode(data=10)),
            ),
        )
        assert parse(expr) == expected

    def test_parse_instance(self):
        expr = {
            "$call": "path/to/a/script.py:ClassName",
            "$args": {
                "a": 10,
                "b": {
                    "$call": "some.interesting.module.MyClass",
                    "$args": {
                        "foo": "hello",
                        "bar": "world",
                    },
                },
            },
        }
        expected = ast.InstanceNode(
            symbol=ast.LiteralNode(data="path/to/a/script.py:ClassName"),
            args=ast.DictNode(
                nodes={
                    ast.LiteralNode(data="a"): ast.LiteralNode(data=10),
                    ast.LiteralNode(data="b"): ast.InstanceNode(
                        symbol=ast.LiteralNode(data="some.interesting.module.MyClass"),
                        args=ast.DictNode(
                            nodes={
                                ast.LiteralNode(data="foo"): ast.LiteralNode(
                                    data="hello"
                                ),
                                ast.LiteralNode(data="bar"): ast.LiteralNode(
                                    data="world"
                                ),
                            }
                        ),
                    ),
                }
            ),
        )
        assert parse(expr) == expected

    def test_parse_model(self):
        expr = {
            "$model": "path/to/a/script.py:ModelName",
            "$args": {
                "a": 10,
                "b": {
                    "foo": "hello",
                    "bar": "world",
                },
            },
        }
        expected = ast.ModelNode(
            symbol=ast.LiteralNode(data="path/to/a/script.py:ModelName"),
            args=ast.DictNode(
                nodes={
                    ast.LiteralNode(data="a"): ast.LiteralNode(data=10),
                    ast.LiteralNode(data="b"): ast.DictNode(
                        nodes={
                            ast.LiteralNode(data="foo"): ast.LiteralNode(data="hello"),
                            ast.LiteralNode(data="bar"): ast.LiteralNode(data="world"),
                        }
                    ),
                }
            ),
        )
        assert parse(expr) == expected

    def test_parse_instance_no_args(self):
        expr = {"$call": "path/to/a/script.py:ClassName"}
        expected = ast.InstanceNode(
            symbol=ast.LiteralNode(data="path/to/a/script.py:ClassName"),
            args=ast.DictNode(nodes={}),
        )
        assert parse(expr) == expected

    def test_parse_model_no_args(self):
        expr = {"$model": "path/to/a/script.py:ModelName"}
        expected = ast.ModelNode(
            symbol=ast.LiteralNode(data="path/to/a/script.py:ModelName"),
            args=ast.DictNode(nodes={}),
        )
        assert parse(expr) == expected

    def test_parse_for(self):
        expr = {"$for(iterable, x)": {"node_$index(x)": "Hello_$item(x)"}}
        expected = ast.ForNode(
            iterable=ast.LiteralNode(data="iterable"),
            body=ast.DictNode(
                nodes={
                    ast.StrBundleNode(
                        ast.LiteralNode(data="node_"),
                        ast.IndexNode(identifier=ast.LiteralNode(data="x")),
                    ): ast.StrBundleNode(
                        ast.LiteralNode(data="Hello_"),
                        ast.ItemNode(identifier=ast.LiteralNode(data="x")),
                    )
                }
            ),
            identifier=ast.LiteralNode(data="x"),
        )
        assert parse(expr) == expected

    def test_parse_for_compact(self):
        expr = {"$for(iterable)": {"node_$index": "Hello_$item"}}
        expected = ast.ForNode(
            iterable=ast.LiteralNode(data="iterable"),
            body=ast.DictNode(
                nodes={
                    ast.StrBundleNode(
                        ast.LiteralNode(data="node_"), ast.IndexNode()
                    ): ast.StrBundleNode(ast.LiteralNode(data="Hello_"), ast.ItemNode())
                }
            ),
        )
        assert parse(expr) == expected

    def test_parse_for_multiple(self):
        expr = {
            "$for(alpha)": {"node_$index": "Hello_$item"},
            "$for(beta)": {"node_$index": "Ciao_$item"},
            "$for(gamma)": {"node_$index": "Hola_$item"},
        }
        expected = ast.DictBundleNode(
            ast.ForNode(
                iterable=ast.LiteralNode(data="alpha"),
                body=ast.DictNode(
                    nodes={
                        ast.StrBundleNode(
                            ast.LiteralNode(data="node_"), ast.IndexNode()
                        ): ast.StrBundleNode(
                            ast.LiteralNode(data="Hello_"), ast.ItemNode()
                        )
                    }
                ),
            ),
            ast.ForNode(
                iterable=ast.LiteralNode(data="beta"),
                body=ast.DictNode(
                    nodes={
                        ast.StrBundleNode(
                            ast.LiteralNode(data="node_"), ast.IndexNode()
                        ): ast.StrBundleNode(
                            ast.LiteralNode(data="Ciao_"), ast.ItemNode()
                        )
                    }
                ),
            ),
            ast.ForNode(
                iterable=ast.LiteralNode(data="gamma"),
                body=ast.DictNode(
                    nodes={
                        ast.StrBundleNode(
                            ast.LiteralNode(data="node_"), ast.IndexNode()
                        ): ast.StrBundleNode(
                            ast.LiteralNode(data="Hola_"), ast.ItemNode()
                        )
                    }
                ),
            ),
        )
        assert parse(expr) == expected

    def test_parse_for_multiple_mixed(self):
        expr = {
            "$for(alpha)": {"node_$index": "Hello_$item"},
            "$for(beta)": {"node_$index": "Ciao_$item"},
            "a": 10,
            "b": {"c": 10.0, "d": "hello"},
        }
        expected = ast.DictBundleNode(
            ast.ForNode(
                iterable=ast.LiteralNode(data="alpha"),
                body=ast.DictNode(
                    nodes={
                        ast.StrBundleNode(
                            ast.LiteralNode(data="node_"), ast.IndexNode()
                        ): ast.StrBundleNode(
                            ast.LiteralNode(data="Hello_"), ast.ItemNode()
                        )
                    }
                ),
            ),
            ast.ForNode(
                iterable=ast.LiteralNode(data="beta"),
                body=ast.DictNode(
                    nodes={
                        ast.StrBundleNode(
                            ast.LiteralNode(data="node_"), ast.IndexNode()
                        ): ast.StrBundleNode(
                            ast.LiteralNode(data="Ciao_"), ast.ItemNode()
                        )
                    }
                ),
            ),
            ast.DictNode(
                nodes={
                    ast.LiteralNode(data="a"): ast.LiteralNode(data=10),
                    ast.LiteralNode(data="b"): ast.DictNode(
                        nodes={
                            ast.LiteralNode(data="c"): ast.LiteralNode(data=10.0),
                            ast.LiteralNode(data="d"): ast.LiteralNode(data="hello"),
                        }
                    ),
                }
            ),
        )
        assert parse(expr) == expected

    def test_parse_switch_base(self):
        expr = {
            "$switch(nation)": [
                {
                    "$case": ["UK", "USA", "Australia"],
                    "$then": "hello",
                },
                {
                    "$case": "Italy",
                    "$then": "ciao",
                },
                {
                    "$default": "*raise your hand*",
                },
            ]
        }
        expected = ast.SwitchNode(
            value=ast.LiteralNode(data="nation"),
            cases=[
                (
                    ast.ListNode(
                        ast.LiteralNode(data="UK"),
                        ast.LiteralNode(data="USA"),
                        ast.LiteralNode(data="Australia"),
                    ),
                    ast.LiteralNode(data="hello"),
                ),
                (
                    ast.LiteralNode(data="Italy"),
                    ast.LiteralNode(data="ciao"),
                ),
            ],
            default=ast.LiteralNode(data="*raise your hand*"),
        )
        print(parse(expr))
        assert parse(expr) == expected


class TestStringParse:
    def test_simple(self):
        expr = "I am a string"
        assert parse(expr) == ast.LiteralNode(data=expr)

    @pytest.mark.parametrize(
        ["id_", "default", "env", "help"],
        [
            ["var1", 10, False, "help"],
            ["var1.var2", "hello", True, ""],
            ["var.var.var", 10.0, True, "help with many words"],
        ],
    )
    def test_var(self, id_: Any, default: Any, env: bool, help: str):
        default_str = f'"{default}"' if isinstance(default, str) else default
        expr = f'$var({id_}, default={default_str}, env={env}, help="{help}")'
        print(expr)
        assert parse(expr) == ast.VarNode(
            identifier=ast.LiteralNode(data=id_),
            default=ast.LiteralNode(data=default),
            env=ast.LiteralNode(data=env),
            help=ast.LiteralNode(data=help),
        )

    def test_import(self):
        path = "path/to/my/file.json"
        expr = f"$import('{path}')"
        assert parse(expr) == ast.ImportNode(path=ast.LiteralNode(data=path))

    def test_sweep(self):
        expr = "$sweep(10, foo.bar, '30')"
        expected = ast.SweepNode(
            ast.LiteralNode(data=10),
            ast.LiteralNode(data="foo.bar"),
            ast.LiteralNode(data="30"),
        )
        assert parse(expr) == expected

    def test_str_bundle(self):
        expr = "I am a string with $var(one.two.three) and $sweep(10, foo.bar, '30')"
        expected = ast.StrBundleNode(
            ast.LiteralNode(data="I am a string with "),
            ast.VarNode(identifier=ast.LiteralNode(data="one.two.three")),
            ast.LiteralNode(data=" and "),
            ast.SweepNode(
                ast.LiteralNode(data=10),
                ast.LiteralNode(data="foo.bar"),
                ast.LiteralNode(data="30"),
            ),
        )
        assert parse(expr) == expected

    def test_uuid(self):
        expr = "$uuid"
        expected = ast.UuidNode()
        assert parse(expr) == expected

    @pytest.mark.parametrize(["format_"], [[None], ["%Y-%m-%d"], ["%h:%M:%s"]])
    def test_date(self, format_):
        expr = "$date"
        if format_ is not None:
            expr += f'("{format_}")'
        format_ = ast.LiteralNode(data=format_) if format_ is not None else None
        assert parse(expr) == ast.DateNode(format=format_)

    @pytest.mark.parametrize(["name"], [[None], ["my_name"]])
    def test_tmp(self, name):
        expr = "$tmp"
        if name is not None:
            expr += f'("{name}")'
        name = ast.LiteralNode(data=name) if name is not None else None
        assert parse(expr) == ast.TmpDirNode(name=name)

    @pytest.mark.parametrize(["symbol"], [["builtins.str"], ["np.array"]])
    def test_symbol(self, symbol: str) -> None:
        expr = f"$symbol({symbol})"
        assert parse(expr) == ast.SymbolNode(symbol=ast.LiteralNode(data=symbol))


class TestParserRaise:
    def test_unknown_directive(self):
        expr = "$unknown_directive(lots, of, params=10)"
        with pytest.raises(ChoixeParsingError):
            parse(expr)

    def test_arg_too_complex(self):
        expr = "$sweep(lots, of, [arguments, '10'])"
        with pytest.raises(ChoixeParsingError):
            parse(expr)

    @pytest.mark.parametrize(
        ["expr"],
        [
            ["$var(f.dd111])"],
            ["I am a string with $import(ba[a)"],
            ["$var(invalid syntax ::) that raises syntaxerror"],
            ["$a+b a"],
        ],
    )
    def test_syntax_error(self, expr: str):
        with pytest.raises(ChoixeParsingError):
            parse(expr)

    def test_var_inline_list_without_quotes(self):
        from yaml import safe_load
        from io import StringIO

        # Never do this btw
        yaml_expr = """a: [ $var(one.two.three, default=10) ]"""

        loaded = safe_load(StringIO(yaml_expr))
        with pytest.raises(ChoixeParsingError):
            parse(loaded)



from pydantic import Field

from pipelime.sequences import Sample, piped_sequence
from pipelime.sequences.pipes import PipedSequenceBase


@piped_sequence
class ReverseSequence(PipedSequenceBase, title="reversed"):
    """Reverses the order of the first `num` samples."""

    num: int = Field(..., description="The number of samples to reverse.")

    def get_sample(self, idx: int) -> Sample:
        if idx < self.num:
            return self.source[self.num - idx - 1]
        return self.source[idx]



from abc import abstractmethod
from enum import Enum
import typing as t

from pydantic import BaseModel, Field, PrivateAttr

from pipelime.piper.progress.tracker.base import Tracker


class PiperPortType(Enum):
    INPUT = "input"
    OUTPUT = "output"
    PARAMETER = "parameter"


class PiperInfo(BaseModel, extra="forbid", copy_on_model_validation="none"):
    token: str = Field("", description="The piper execution token.")
    node: str = Field("", description="The piper dag's node name.")

    @property
    def active(self) -> bool:
        return bool(self.token)


class PipelimeCommand(
    BaseModel,
    allow_population_by_field_name=True,
    extra="forbid",
    copy_on_model_validation="none",
):
    """Base class for all pipelime commands.
    Subclasses should implement the run method.
    """

    _piper: PiperInfo = PrivateAttr(default_factory=PiperInfo)  # type: ignore
    _tracker: t.Optional[Tracker] = PrivateAttr(None)

    @abstractmethod
    def run(self) -> None:
        pass

    @classmethod
    def _filter_fields_by_flag(cls, flag: str, value: t.Any) -> t.Iterable[str]:
        for k, v in cls.__fields__.items():
            if v.field_info.extra.get(flag, object()) == value:
                yield k

    def _get_fields_by_flag(self, flag: str, value: t.Any) -> t.Dict[str, t.Any]:
        return {k: getattr(self, k) for k in self._filter_fields_by_flag(flag, value)}

    def get_inputs(self) -> t.Dict[str, t.Any]:
        return self._get_fields_by_flag("piper_port", PiperPortType.INPUT)

    def get_outputs(self) -> t.Dict[str, t.Any]:
        return self._get_fields_by_flag("piper_port", PiperPortType.OUTPUT)

    def _get_piper_tracker(self) -> Tracker:
        if self._tracker is None:
            from pipelime.piper.progress.tracker.factory import TrackCallbackFactory

            cb = TrackCallbackFactory.get_callback()
            self._tracker = Tracker(self._piper.token, self._piper.node, cb)

        return self._tracker

    @classmethod
    def command_title(cls) -> str:
        if cls.__config__.title:
            return cls.__config__.title
        return cls.__name__

    def command_name(self) -> str:
        return self.command_title()

    def set_piper_info(
        self,
        *,
        token: t.Optional[str] = None,
        node: t.Optional[str] = None,
    ):
        if token is not None:
            self._piper.token = token
        if node is not None:
            self._piper.node = node

    def track(
        self,
        seq: t.Union[t.Sequence, t.Iterable],
        *,
        size: t.Optional[int] = None,
        message: str = "",
    ) -> t.Iterable:
        import rich.progress

        if self._piper.active:
            tracker = self._get_piper_tracker()
            return tracker.track(seq, size=size, message=message)
        else:
            return rich.progress.track(
                seq,
                total=len(seq) if size is None else size,  # type: ignore
                description="🍋 " + message,
            )

    def __call__(self) -> None:
        self.run()


class DAGModel(BaseModel, extra="forbid", copy_on_model_validation="none"):
    """A Piper DAG as a `<node>: <command>` mapping."""

    nodes: t.Mapping[str, PipelimeCommand]

    def purged_dict(self):
        from pipelime.choixe import XConfig

        return XConfig(data={"nodes": self.nodes}).decode()



import typing as t
from pathlib import Path

import pydantic as pyd
from filelock import FileLock, Timeout
from tempfile import TemporaryDirectory

import pipelime.sequences as pls
from pipelime.items import Item, SerializationMode
from pipelime.sequences.pipes import PipedSequenceBase


class _serialization_mode_override:
    """Changes the serialization mode of a specific item."""

    def __init__(self, item: Item, mode: t.Optional[t.Union[SerializationMode, str]]):
        self._item = item
        self._mode = mode

    def __enter__(self):
        if self._mode is not None:
            self._prev_mode = self._item.serialization_mode
            self._item.serialization_mode = self._mode

    def __exit__(self, exc_type, exc_value, traceback):
        if self._mode is not None:
            self._item.serialization_mode = self._prev_mode


@pls.piped_sequence
class UnderfolderWriter(
    PipedSequenceBase, title="to_underfolder", underscore_attrs_are_private=True
):
    """Writes samples to an underfolder dataset while iterating over them."""

    folder: Path = pyd.Field(..., description="The output folder.")
    zfill: t.Optional[int] = pyd.Field(None, description="Custom index zero-filling.")
    key_serialization_mode: t.Optional[
        t.Mapping[str, t.Union[SerializationMode, str]]
    ] = pyd.Field(None, description="Forced serialization mode for each key.")
    exists_ok: bool = pyd.Field(
        False, description="If False raises an error when `folder` exists."
    )

    _data_folder: Path
    _effective_zfill: int
    _temp_folder: TemporaryDirectory

    def __init__(self, folder: Path, **data):
        super().__init__(folder=folder, **data)  # type: ignore

        self._data_folder = self.folder / "data"
        self._effective_zfill = (
            self.source.best_zfill() if self.zfill is None else self.zfill
        )
        if self.key_serialization_mode is None:
            self.key_serialization_mode = {}

        if not self.exists_ok and (self.folder.exists() or self._data_folder.exists()):
            raise FileExistsError(
                "Trying to overwrite an existing dataset: `"
                + str(self.folder)
                + "`. Please use `exists_ok=True` to overwrite."
            )

        self._data_folder.mkdir(parents=True, exist_ok=True)
        self._temp_folder = TemporaryDirectory()  # will be automatically deleted

    def get_sample(self, idx: int) -> pls.Sample:
        sample = self.source[idx]

        id_str = str(idx)
        id_str = id_str.zfill(self._effective_zfill)

        for k, v in sample.items():
            with _serialization_mode_override(
                v, self.key_serialization_mode.get(k)  # type: ignore
            ):
                if v.is_shared:
                    filepath = self.folder / k
                    if not any(f.exists() for f in v.get_all_names(filepath)):
                        lock = FileLock(
                            str(Path(self._temp_folder.name) / (k + ".~lock"))
                        )
                        try:
                            with lock.acquire(timeout=1):
                                # check again to avoid races
                                if not any(
                                    f.exists() for f in v.get_all_names(filepath)
                                ):
                                    v.serialize(filepath)
                        except Timeout:  # pragma: no cover
                            pass
                else:
                    v.serialize(self._data_folder / f"{id_str}_{k}")

        return sample



import typing as t
from pathlib import Path

import pydantic as pyd

import pipelime.sequences as pls
from pipelime.utils.pydantic_types import ItemType
from pipelime.stages.base import SampleStage, StageInput
from pipelime.stages import StageKeyFormat
from pipelime.sequences.pipes import PipedSequenceBase


@pls.piped_sequence
class MappedSequence(PipedSequenceBase, title="map"):
    """Applies a stage on all samples."""

    stage: StageInput = pyd.Field(..., description=StageInput.__doc__)  # type: ignore

    def __init__(
        self,
        stage: t.Union[SampleStage, str, t.Mapping[str, t.Mapping[str, t.Any]]],
        **data,
    ):
        super().__init__(stage=stage, **data)  # type: ignore

    def get_sample(self, idx: int) -> pls.Sample:
        return self.stage(self.source[idx])


@pls.piped_sequence
class ZippedSequences(PipedSequenceBase, title="zip"):
    """Zips two Sequences by merging each Sample."""

    to_zip: pls.SamplesSequence = pyd.Field(..., description="The sequence to merge.")
    key_format: str = pyd.Field(
        "*",
        description=(
            "The zipped samples' key format. Any `*` will be replaced with the "
            "source key, eg, `my_*_key` on [`image`, `mask`] generates "
            "`my_image_key` and `my_mask_key`. If no `*` is found, the string is "
            "suffixed to source key, ie, `MyKey` on `image` gives "
            "`imageMyKey`. If empty, the source key will be used as-is."
        ),
    )

    _key_formatting_stage: StageKeyFormat = pyd.PrivateAttr()

    @pyd.validator("key_format")
    def validate_key_format(cls, v):
        if "*" in v:
            return v
        return "*" + v

    def __init__(self, to_zip: pls.SamplesSequence, **data):
        super().__init__(to_zip=to_zip, **data)  # type: ignore
        self._key_formatting_stage = StageKeyFormat(key_format=self.key_format)

    def size(self) -> int:
        return min(len(self.source), len(self.to_zip))

    def get_sample(self, idx: int) -> pls.Sample:
        return self.source[idx].merge(self._key_formatting_stage(self.to_zip[idx]))


@pls.piped_sequence
class ConcatSequences(PipedSequenceBase, title="cat"):
    """Concatenates two SamplesSequences."""

    to_cat: pls.SamplesSequence = pyd.Field(
        ..., description="The samples to concatenate."
    )

    def __init__(self, to_cat: pls.SamplesSequence, **data):
        super().__init__(to_cat=to_cat, **data)  # type: ignore

    def size(self) -> int:
        return len(self.source) + len(self.to_cat)

    def get_sample(self, idx: int) -> pls.Sample:
        if idx < len(self.source):
            return self.source[idx]
        return self.to_cat[idx - len(self.source)]


@pls.piped_sequence
class FilteredSequence(
    PipedSequenceBase, title="filter", underscore_attrs_are_private=True
):
    """A filtered view of a SamplesSequence."""

    filter_fn: t.Callable[[pls.Sample], bool] = pyd.Field(
        ..., description="A callable returning True for any valid sample."
    )

    _valid_idxs: t.Optional[t.Sequence[int]] = None

    def __init__(self, filter_fn: t.Callable[[pls.Sample], bool], **data):
        super().__init__(filter_fn=filter_fn, **data)  # type: ignore

    def _get_valid_idxs(self) -> t.Sequence[int]:
        if self._valid_idxs is None:
            self._valid_idxs = [
                idx for idx, sample in enumerate(self.source) if self.filter_fn(sample)
            ]
        return self._valid_idxs

    def size(self) -> int:
        return len(self._get_valid_idxs())

    def get_sample(self, idx: int) -> pls.Sample:
        return self.source[self._get_valid_idxs()[idx]]

    def __iter__(self) -> t.Iterator[pls.Sample]:
        _src_idx = 0
        while _src_idx < len(self.source):
            x = self.source[_src_idx]
            if self.filter_fn(x):
                yield x
            _src_idx += 1


@pls.piped_sequence
class SortedSequence(
    PipedSequenceBase, title="sort", underscore_attrs_are_private=True
):
    """A sorted view of an input SamplesSequence."""

    key_fn: t.Callable[[pls.Sample], t.Any] = pyd.Field(
        ...,
        description=(
            "The key function to compare Samples. Use `functools.cmp_to_key` to "
            "convert a compare function, ie, accepting two arguments, to a key "
            "function."
        ),
    )

    _sorted_idxs: t.Sequence[int]

    def __init__(self, key_fn: t.Callable[[pls.Sample], t.Any], **data):
        super().__init__(key_fn=key_fn, **data)  # type: ignore
        self._sorted_idxs = sorted(
            range(len(self.source)), key=lambda k: self.key_fn(self.source[k])
        )

    def get_sample(self, idx: int) -> pls.Sample:
        return self.source[self._sorted_idxs[idx]]


@pls.piped_sequence
class SlicedSequence(
    PipedSequenceBase, title="slice", underscore_attrs_are_private=True
):
    """Extracts a slice [start_idx:end_idx:step] from the input SamplesSequence."""

    start: t.Optional[int] = pyd.Field(
        None, description="The first index, defaults to the first element."
    )
    stop: t.Optional[int] = pyd.Field(
        None, description="The final index, defaults to the last element."
    )
    step: t.Optional[int] = pyd.Field(
        None, description="The slice step, defaults to 1."
    )

    _sliced_idxs: t.Sequence[int]

    def __init__(self, **data):
        super().__init__(**data)

        effective_start = (
            0
            if self.start is None
            else max(0, min(len(self.source), self._normalized(self.start)))
        )
        effective_stop = (
            len(self.source)
            if self.stop is None
            else max(0, min(len(self.source), self._normalized(self.stop)))
        )
        effective_step = 1 if self.step is None else self.step
        self._sliced_idxs = range(effective_start, effective_stop, effective_step)

    def _normalized(self, idx: int) -> int:
        return idx if idx >= 0 else len(self.source) + idx

    def size(self) -> int:
        return len(self._sliced_idxs)

    def get_sample(self, idx: int) -> pls.Sample:
        return self.source[self._sliced_idxs[idx]]


@pls.piped_sequence
class IndexSelectionSequence(
    PipedSequenceBase, title="select", underscore_attrs_are_private=True
):
    """Given a list of indexes, extracts the corresponding samples from the input
    SamplesSequence. The index sequence is not automatically sorted.
    """

    indexes: t.Sequence[int] = pyd.Field(
        ...,
        description=(
            "The indexes to extract. Negative values start counting from the end."
        ),
    )
    negate: bool = pyd.Field(
        False,
        description=(
            "FALSE to extract samples at `indexes`, "
            "TRUE to get all samples but the ones at `indexes`."
        ),
    )

    _effective_idxs: t.Sequence[int]

    def __init__(self, indexes: t.Sequence[int], **data):
        super().__init__(indexes=indexes, **data)  # type: ignore

        for idx in self.indexes:
            if idx < 0:
                idx += len(self.source)
            if idx < 0 or idx >= len(self.source):
                raise ValueError(
                    "Index {} is out of range [0, {})".format(idx, len(self.source))
                )
        self._effective_idxs = (
            [idx for idx in range(len(self.source)) if idx not in self.indexes]
            if self.negate
            else self.indexes
        )

    def size(self) -> int:
        return len(self._effective_idxs)

    def get_sample(self, idx: int) -> pls.Sample:
        return self.source[self._effective_idxs[idx]]


@pls.piped_sequence
class ShuffledSequence(
    PipedSequenceBase, title="shuffle", underscore_attrs_are_private=True
):
    """Shuffles samples in the input SamplesSequence."""

    seed: t.Optional[int] = pyd.Field(None, description="The optional random seed.")

    _shuffled_idxs: t.Sequence[int]

    def __init__(self, **data):
        import random

        super().__init__(**data)
        if self.seed is not None:
            random.seed(self.seed)
        self._shuffled_idxs = list(range(len(self.source)))
        random.shuffle(self._shuffled_idxs)

    def get_sample(self, idx: int) -> pls.Sample:
        return self.source[self._shuffled_idxs[idx]]


@pls.piped_sequence
class EnumeratedSequence(
    PipedSequenceBase, title="enumerate", underscore_attrs_are_private=True
):
    """Adds a new index item to each Sample in the input SamplesSequence."""

    idx_key: str = pyd.Field(
        "~idx", description="The new key containing the index item."
    )
    item_cls: ItemType = pyd.Field(
        ItemType.make_default("TxtNumpyItem"),
        description="The item class holding the index.",
    )

    def get_sample(self, idx: int) -> pls.Sample:
        sample = self.source[idx]
        return sample.set_item(
            key=self.idx_key, value=self.item_cls(idx)  # type: ignore
        )


@pls.piped_sequence
class RepeatedSequence(PipedSequenceBase, title="repeat"):
    """Repeats this sequence so each sample is seen multiple times."""

    count_: pyd.NonNegativeInt = pyd.Field(
        ..., alias="count", description="The number of repetition."
    )

    def __init__(self, count: pyd.NonNegativeInt, **data):
        super().__init__(count=count, **data)  # type: ignore

    def size(self) -> int:
        return len(self.source) * self.count_

    def get_sample(self, idx: int) -> pls.Sample:
        if idx < 0 or idx >= len(self):
            raise IndexError(f"Sample index `{idx}` is out of range.")
        return self.source[idx % len(self.source)]


@pls.piped_sequence
class CachedSequence(PipedSequenceBase, title="cache"):
    """Caches the input Samples the first time they are accessed."""

    cache_folder: t.Optional[Path] = pyd.Field(
        None,
        description=(
            "The cache folder path. Leave empty to use a temporary folder "
            "which will be deleted when exiting the process."
        ),
    )
    reuse_cache: bool = pyd.Field(
        False,
        description=(
            "If `cache_folder` exists, use it anyway if true, "
            "otherwise raise a FileExistsError."
        ),
    )

    _temp_folder = pyd.PrivateAttr(None)

    def __init__(self, cache_folder: t.Optional[Path] = None, **data):
        super().__init__(cache_folder=cache_folder, **data)  # type: ignore
        if self.cache_folder is None:
            from tempfile import TemporaryDirectory

            self._temp_folder = TemporaryDirectory()
            self.cache_folder = Path(self._temp_folder.name)
        else:
            if not self.reuse_cache and self.cache_folder.exists():
                raise FileExistsError(
                    f"The cache folder `{self.cache_folder}` already exists."
                )
            self.cache_folder.mkdir(parents=True, exist_ok=True)

    def get_sample(self, idx: int) -> pls.Sample:
        import pickle

        filename: Path = self.cache_folder / f"{idx}.pkl"  # type: ignore

        if filename.exists():
            with open(filename, "rb") as fd:
                return pickle.load(fd)

        x = self.source[idx]
        self._cache_sample(x, filename)
        return x

    def _cache_sample(self, x: pls.Sample, filename: Path):
        from filelock import FileLock, Timeout
        import pickle

        lock = FileLock(str(filename.with_suffix(filename.suffix + ".~lock")))
        try:
            with lock.acquire(timeout=1):
                # check again to avoid races
                if not filename.exists():
                    with open(filename, "wb") as fd:
                        pickle.dump(x, fd, protocol=-1)
        except Timeout:  # pragma: no cover
            pass


@pls.piped_sequence
class EnableItemDataCache(PipedSequenceBase, title="data_cache"):
    """Enables item data caching on previous pipeline steps."""

    items: t.Union[ItemType, t.Sequence[ItemType]] = pyd.Field(
        default_factory=list,
        description="One or more item classes where data cache should be enabled.",
    )

    _item_cls: list = pyd.PrivateAttr()

    def __init__(self, *items, **data):
        super().__init__(items=items, **data)  # type: ignore
        self._item_cls = [
            it.itype
            for it in (
                self.items if isinstance(self.items, t.Sequence) else [self.items]
            )
        ]

    def get_sample(self, idx: int) -> pls.Sample:
        from pipelime.items import data_cache

        with data_cache(*self._item_cls):  # type: ignore
            return super().get_sample(idx)


@pls.piped_sequence
class DisableItemDataCache(PipedSequenceBase, title="no_data_cache"):
    """Disables item data caching on previous pipeline steps."""

    items: t.Union[ItemType, t.Sequence[ItemType]] = pyd.Field(
        default_factory=list,
        description="One or more item classes where data cache should be disabled.",
    )

    _item_cls: list = pyd.PrivateAttr()

    def __init__(self, *items, **data):
        super().__init__(items=items, **data)  # type: ignore
        self._item_cls = [
            it.itype
            for it in (
                self.items if isinstance(self.items, t.Sequence) else [self.items]
            )
        ]

    def get_sample(self, idx: int) -> pls.Sample:
        from pipelime.items import no_data_cache

        with no_data_cache(*self._item_cls):  # type: ignore
            return super().get_sample(idx)



from pipelime.stages.base import SampleStage, StageCompose, StageIdentity, StageLambda
from pipelime.stages.augmentations import StageAlbumentations
from pipelime.stages.item_replacement import StageReplaceItem
from pipelime.stages.item_sources import StageForgetSource, StageUploadToRemote
from pipelime.stages.key_transformations import (
    StageDuplicateKey,
    StageKeyFormat,
    StageKeysFilter,
    StageRemap,
)
from pipelime.stages.item_info import StageItemInfo
